;; Generated package description from vecdb.el  -*- mode: lisp-data; no-byte-compile: t -*-
(define-package "vecdb" "0.2.0.20250721.13325" "An interface to vector databases for embeddings" '((emacs "29.1") (plz "0.8") (pg "0.56")) :commit "e89d5a9a3abb6b1c0a72dbe2849ee8193a3c4019" :authors '(("Andrew Hyatt" . "ahyatt@gmail.com")) :maintainer '("Andrew Hyatt" . "ahyatt@gmail.com") :url "https://github.com/ahyatt/vecdb")
